/*
**	AIFF DataType
**
**	Written by Olaf `Olsen' Barthel <olsen@sourcery.han.de>
**		Public domain
**
**	:ts=4
*/

#include "Data.h"

/****************************************************************************/

#define FAILED (FALSE)

/****************************************************************************/

	// Maximum supported replay rate, as per the "Amiga Hardware Reference Manual"

#define MAX_SAMPLE_RATE 28867

	// The minimum sample rate we will allow when scaling a sound down

#define MIN_SAMPLE_RATE 5563

	// How many bytes to read in one piece

#define MIN_FRAME_RATE	(2048 * 8)

	// 80 bit IEEE Standard 754 floating point number

typedef struct {
	unsigned short	exponent;			// Exponent, bit #15 is sign bit for mantissa
	unsigned long	mantissa[2];		// 64 bit mantissa
} extended;

	// Audio Interchange Format chunk data

#define ID_AIFF MAKE_ID('A','I','F','F')
#define ID_AIFC MAKE_ID('A','I','F','C')

#define ID_FVER MAKE_ID('F','V','E','R')
#define ID_COMM MAKE_ID('C','O','M','M')
#define ID_SSND MAKE_ID('S','S','N','D')

	// "COMM" chunk header

typedef struct {
	short			numChannels;		// Number of channels
	unsigned long	numSampleFrames;	// Number of sample frames
	short			sampleSize; 		// Number of bits per sample point
	extended		sampleRate; 		// Replay rate in samples per second
} CommonChunk;

	// The same for "AIFC" type files; this should be longer, but we don't
	// need the name of the compression format

typedef struct {
	short			numChannels;		// Number of channels
	unsigned long	numSampleFrames;	// Number of sample frames
	short			sampleSize; 		// Number of bits per sample point
	extended		sampleRate; 		// Replay rate in samples per second
	unsigned long	compressionType;	// Compression type
} ExtCommonChunk;

#define NO_COMPRESSION MAKE_ID('N','O','N','E') // No sound compression

	// "SSND" chunk header

typedef struct {
	unsigned long	offset, 			// Offset to sound data, for block alignment
					blockSize;			// Size of block data is aligned to
} SampledSoundHeader;

	// "FVER" chunk header

typedef struct {
	long			timestamp;			// Format version creation date
} FormatVersionHeader;

#define AIFCVersion1 0xA2805140 		// "AIFC" file format version #1

/****************************************************************************/

	/* extended2long(extended *ex):
	 *
	 *	Convert an 80 bit IEEE Standard 754 floating point number
	 *	into an integer value.
	 */

static long
extended2long(extended *ex)
{
	unsigned long mantissa;
	short exponent;
	long sign;

		// We only need 32 bits precision

	mantissa = ex->mantissa[0];

		// Is the mantissa positive or negative?

	exponent = ex->exponent;

	if(exponent & (1 << 15))
		sign = -1;
	else
		sign =	1;

		// Unbias the exponent (strip the sign bit; the
		// exponent is 15 bits wide)

	exponent = (exponent & ~(1 << 15)) - ((1 << 14) - 1);

		// If the exponent is negative, set the mantissa to zero.
		// We cannot represent integer values >0 and <1.

	if(exponent < 0)
		mantissa = 0;
	else
	{
			// We used only the upper 32 bits of the mantissa,
			// which is what we have to make up for. Subtracting
			// 31 from the exponent is actually dividing by
			// 2^32 (i.e. x / (1L<<31)).

		exponent -= 31;

			// If the exponent is not negative, then the value
			// the number represents will be larger than the
			// original 64 bits of the mantissa would hold.

		if(exponent > 0)
			mantissa = (1L << 31) - 1;	// == MAXINT
		else
			mantissa >>= -exponent;		// Keep the integer part of the number
	}

		// That's all...

	return(sign * (long)mantissa);
}

	/* Flatten():
	 *
	 *	Flatten the "AIFF"/"AIFC" sound data, i.e. blend all the channels
	 *	into a single monophonic chunk and crop the data to eight bits per
	 *	sample.
	 */

STATIC VOID
Flatten(
	UBYTE *		Src,			// Sample data
	BYTE *		Dst,			// Where to store the processed data
	LONG		BytesPerPoint,	// Number of bytes that make up a single sample
	LONG		NumChannels,	// Number of channels in this sample data
	LONG		NumFrames,		// Number of sample frames in this sample data
	LONG		SamplesToBlend)	// Number of samples to blend into a single sample
{
	LONG Sum,Value,SamplesBlended,FrameIndex,i,j;

		// Convert all the frames if possible

	SamplesBlended = 0;

	for(FrameIndex = 0 ; FrameIndex < NumFrames ; FrameIndex++)
	{
			// Add up the channel data

		Sum = 0;

		for(i = 0; i < NumChannels ; i++)
		{
				// The audio data is always stored left adjusted.

			Value = 0;

			for(j = 0 ; j < BytesPerPoint ; j++)
				Value |= (LONG)Src[j] << ((3 - j) * 8);

				// Move up to the next sample

			Src += BytesPerPoint;

				// Add the new sample, 16 bits only

			Sum += Value / (1L << 16);
		}

			// We store this sample value if it is the
			// very first one or if enough samples
			// have been merged already. Note that
			// SamplesToBlend > 0.

		SamplesBlended++;
		if(SamplesBlended == SamplesToBlend || FrameIndex == 0)
		{
				// Calculate the monophonic sample data and
				// crop it to eight bits

			Sum = (Sum / NumChannels) / (1L << 8);

				// Make sure that the value is in range

			if(Sum < -128)
				Sum = -128;
			else if(Sum > 127)
				Sum = 127;

				// Keep this sample

			*Dst++ = (BYTE)Sum;

				// Blend the next samples if necessary

			SamplesBlended = 0;
		}
	}
}

	/* ConvertAIFF():
	 *
	 *	Convert "AIFF"/"AIFC" audio data into plain 8 bit audio data.
	 */

STATIC BOOL
ConvertAIFF(
	struct IFFHandle *		IFFHandle,
	BYTE **					DstPtr,
	struct VoiceHeader *	VoiceHeader,
	ULONG					MemFlags,
	LONG *					Error,
	struct ClassBase *		ClassBase)
{
	STATIC LONG Stops[] =
	{
		ID_AIFF,ID_COMM,	// AIFF chunks
		ID_AIFF,ID_SSND,

		ID_AIFC,ID_FVER,	// AIFC chunks
		ID_AIFC,ID_COMM,
		ID_AIFC,ID_SSND
	};

	LONG SrcBufferSize,SrcFrames,DstFrames,DstRate,DstSkip,Size;
	FormatVersionHeader FormatHeader;
	SampledSoundHeader SampleHeader;
	ExtCommonChunk Common;
	UBYTE *SrcBuffer;
	BYTE *DstBuffer;
	BOOL Result;

	SrcBuffer	= NULL;
	DstBuffer	= NULL;
	Result 		= FALSE;

		// There are only two/three mandatory chunks to look for

	if(!(*Error = StopChunks(IFFHandle,Stops,5)))
	{
		struct ContextNode *Chunk;
		LONG BytesPerPoint;

		while(!Result && !(*Error) && !ParseIFF(IFFHandle,IFFPARSE_SCAN))
		{
			Chunk = CurrentChunk(IFFHandle);

			switch(Chunk->cn_ID)
			{
					// This is the file format version ID

				case ID_FVER:

					if(ReadChunkBytes(IFFHandle,&FormatHeader,sizeof(FormatVersionHeader)) != sizeof(FormatVersionHeader))
					{
						*Error = IFFERR_READ;

						break;
					}

						// Does this reader support this format?

					if(FormatHeader.timestamp != AIFCVersion1)
					{
						*Error = ERROR_NOT_IMPLEMENTED;

						break;
					}

					break;

					// This chunk is common for all "AIFF" variants

				case ID_COMM:

						// Determine how many bytes to read

					if(Chunk->cn_Type == ID_AIFF)
						Size = sizeof(CommonChunk);
					else
						Size = sizeof(ExtCommonChunk);

					if(ReadChunkBytes(IFFHandle,&Common,Size) != Size)
					{
						*Error = IFFERR_READ;

						break;
					}

						// Is this a compressed "AIFC" file?

					if(Chunk->cn_Type == ID_AIFC && Common.compressionType != NO_COMPRESSION)
					{
						*Error = DTERROR_UNKNOWN_COMPRESSION;

						break;
					}

						// Keep the basic data

					DstFrames	= Common.numSampleFrames;
					DstRate 	= extended2long(&Common.sampleRate);
					DstSkip 	= 1;

						// The current Amiga audio hardware has a fixed
						// replay speed limit. We will compensate for it
						// by cropping the audio data if necessary.

					while(DstRate > MAX_SAMPLE_RATE)
					{
						DstFrames = (DstFrames + 1) / 2;

						DstRate 	/= 2;
						DstSkip 	*= 2;
					}

						// See how many bytes make up one sample point

					if(Common.sampleSize <= 8)
						BytesPerPoint = 1;
					else
					{
						if(Common.sampleSize <= 16)
							BytesPerPoint = 2;
						else
						{
							if(Common.sampleSize <= 24)
								BytesPerPoint = 3;
							else
								BytesPerPoint = 4;
						}
					}

						// Just to be sure we accept multiple "COMM" chunks

					if(SrcBuffer)
						FreeVec(SrcBuffer);

					if(DstBuffer)
					{
						FreeVec(DstBuffer);

						DstBuffer = NULL;
					}

						// Don't waste too much memory for loading

					if(Common.numSampleFrames < MIN_FRAME_RATE)
						SrcFrames = Common.numSampleFrames;
					else
						SrcFrames = MIN_FRAME_RATE;

						// Allocate the decoding buffer

					SrcBufferSize = BytesPerPoint * Common.numChannels * SrcFrames;
					if(!(SrcBuffer = AllocVec(SrcBufferSize,MEMF_ANY)))
					{
						*Error = ERROR_NO_FREE_STORE;

						break;
					}

						// Step down in size in order to make the allocation fit

					while(DstRate > MIN_SAMPLE_RATE && DstFrames > 0 &&
					      (DstBuffer = AllocVec(DstFrames,MemFlags | MEMF_CLEAR)) == NULL)
					{
						DstFrames = (DstFrames + 1) / 2;

						DstRate 	/= 2;
						DstSkip 	*= 2;
					}

						// Any success?

					if(DstBuffer == NULL)
						*Error = ERROR_NO_FREE_STORE;

					break;

					// Here follows the sampled sound; this chunk is somewhat
					// equivalent to the "8SVX"/"BODY" chunk.

				case ID_SSND:

						// Read the data header

					if(ReadChunkBytes(IFFHandle,&SampleHeader,sizeof(SampledSoundHeader)) != sizeof(SampledSoundHeader))
						*Error = IFFERR_READ;
					else
					{
							// Is the data block aligned?

						if(SampleHeader.offset)
						{
							LONG Needed = SampleHeader.offset,Skip;

								// Skip the padding data

							while(!(*Error) && Needed > 0)
							{
								if(SrcBufferSize > Needed)
									Skip = Needed;
								else
									Skip = SrcBufferSize;

								if(ReadChunkBytes(IFFHandle,SrcBuffer,Skip) == Skip)
									Needed -= Skip;
								else
									*Error = IFFERR_READ;
							}
						}

						if(!(*Error))
						{
							LONG FrameSize,TotalFrames,Frames;

							FrameSize	= BytesPerPoint * Common.numChannels;
							TotalFrames	= Common.numSampleFrames;

								// Initialize the voice header

							memset(VoiceHeader,0,sizeof(struct VoiceHeader));

							VoiceHeader->vh_OneShotHiSamples	= DstFrames;
							VoiceHeader->vh_SamplesPerSec 		= DstRate;
							VoiceHeader->vh_Octaves				= 1;
							VoiceHeader->vh_Compression			= CMP_NONE;
							VoiceHeader->vh_Volume				= 64;

							*DstPtr = DstBuffer;

							if(!(*Error))
							{
								BYTE *Dst;

								Dst = DstBuffer;

									// Read the audio data frame by frame

								while(!(*Error) && TotalFrames > 0)
								{
									if(SrcFrames > TotalFrames)
										Frames = TotalFrames;
									else
										Frames = SrcFrames;

									if(ReadChunkRecords(IFFHandle,SrcBuffer,FrameSize,Frames) == Frames)
									{
										Flatten(SrcBuffer,Dst,BytesPerPoint,Common.numChannels,Frames,DstSkip);

										Dst += Frames / DstSkip;

										TotalFrames -= Frames;
									}
									else
										*Error = IFFERR_READ;
								}

								Result = TRUE;
							}
						}
					}

					break;
			}
		}
	}

		// Clean up...

	if(SrcBuffer)
		FreeVec(SrcBuffer);

	if(DstBuffer && (*Error || !Result))
		FreeVec(DstBuffer);

	if(*Error)
		Result = FALSE;

	return(Result);
}

	/* GetAIFF(Object *object,struct TagItem *Tags,struct ClassBase *ClassBase):
	 *
	 *	Create a datatypes object from an "AIFF"/"AIFC" file suitable for
	 *	MultiView, etc. to display or replay.
	 */

STATIC BOOL
GetAIFF(
	Object *			object,
	struct TagItem *	Tags,
	struct ClassBase *	ClassBase)
{
	struct VoiceHeader *	VoiceHeader	= NULL;
	struct IFFHandle *		Handle		= NULL;
	LONG					SourceType	= -1;
	BOOL					Result		= FALSE;
	LONG					Error		= 0;
	STRPTR					FileName;

		// Initialize the file name

	FileName = (STRPTR)GetTagData(DTA_Name,NULL,Tags);

		// Get the basic data
		// NOTE: we assume that the handle we receive
		//       will be an IFFHandle. This is true only
		//       because the datatype descriptor tells the
		//       datatypes.library that this is an IFF
		//       format file. For other file formats you
		//       would have to open the file on your own,
		//       using the file name obtained above.

	GetDTAttrs(object,
		SDTA_VoiceHeader,	&VoiceHeader,
		DTA_Handle, 		&Handle,
		DTA_SourceType,		&SourceType,
	TAG_DONE);

		// Do we have everything we need?

	if(Handle != NULL && VoiceHeader != NULL && FileName != NULL &&
	   (SourceType == DTST_FILE || SourceType == DTST_CLIPBOARD))
	{
		BYTE *Sample;
		ULONG Memory;

			// sound.datatype v40 no longer requires
			// the entire sample to reside in chip memory

		if(SuperClassBase->lib_Version > 39)
			Memory = MEMF_ANY;
		else
			Memory = MEMF_CHIP;

			// Convert the audio file

		if(ConvertAIFF(Handle,&Sample,VoiceHeader,Memory,&Error,ClassBase))
		{
				// Fill in the remaining information

			SetDTAttrs(object,NULL,NULL,
				DTA_ObjName,		FilePart(FileName),
				SDTA_Sample,		Sample,
				SDTA_SampleLength,	VoiceHeader->vh_OneShotHiSamples,
				SDTA_Period,		(ULONG)(SysBase->ex_EClockFrequency * 5) / (ULONG)VoiceHeader->vh_SamplesPerSec,
				SDTA_Volume,		64,
				SDTA_Cycles,		1,
			TAG_DONE);

			Result = TRUE;
		}
	}
	else
	{
		Error = ERROR_OBJECT_NOT_FOUND;
	}

	if(Error != 0)
		SetIoErr(Error);

	return(Result);
}

	/* ClassDispatch():
	 *
	 *	The class dispatcher routine.
	 */

STATIC Object * LIBENT
ClassDispatch(
	REG(a0) Class *		class,
	REG(a2) Object *	object,
	REG(a1) Msg			msg)
{
	struct ClassBase *ClassBase = (struct ClassBase *)class->cl_UserData;
	Object *Result;

		// What message is it?

	switch(msg->MethodID)
	{
			// Create a new instance

		case OM_NEW:

			Result = (Object *)DoSuperMethodA(class,object,msg);
			if(Result != NULL)
			{
				if(GetAIFF(Result,((struct opSet *)msg)->ops_AttrList,ClassBase) == FAILED)
				{
					CoerceMethod(class,Result,OM_DISPOSE);

					Result = NULL;
				}
			}

			break;

			// Let the superclass handle the rest

		default:

			Result = (Object *)DoSuperMethodA(class,object,msg);
			break;
	}

	return(Result);
}

/****************************************************************************/

	/* CreateClass(struct ClassBase *ClassBase):
	 *
	 *	Create the new class and makes it public.
	 */

BOOL
CreateClass(struct ClassBase *ClassBase)
{
	BOOL Result = FALSE;

		// We need the super class for version checking

	SuperClassBase = OpenLibrary("datatypes/sound.datatype",39);
	if(SuperClassBase != NULL)
	{
			// Create the new class

		SoundClass = MakeClass(ClassBase->LibNode.lib_Node.ln_Name,
		                       SOUNDDTCLASS,NULL,0,NULL);

		if(SoundClass != NULL)
		{
			Result = TRUE;

				// Link the class dispatcher into it
				// and keep a pointer to the library
				// base

			SoundClass->cl_Dispatcher.h_Entry	= (HOOKFUNC)ClassDispatch;
			SoundClass->cl_UserData				= (ULONG)ClassBase;

				// Make the class publicly available

			AddClass(SoundClass);
		}
		else
		{
			CloseLibrary(SuperClassBase);
			SuperClassBase = NULL;
		}
	}

	return(Result);
}

	/* GetClassPointer(struct ClassBase *ClassBase):
	 *
	 *	Return a pointer to the class base this
	 *	library implements.
	 */

Class *
GetClassPointer(struct ClassBase *ClassBase)
{
	return(SoundClass);
}

	/* AttemptFreeClass(struct ClassBase *ClassBase):
	 *
	 *	Try to remove the class from the public list.
	 *	Note that when this function is called, it
	 *	should always succeed. If it does not, a user
	 *	of this class must have made a programming error.
	 */

BOOL
AttemptFreeClass(struct ClassBase *ClassBase)
{
	BOOL Result = TRUE;

		// This routine will return success even
		// if the class has already been removed.

	if(SoundClass != NULL)
	{
			// Attempt to remove the class from
			// the public list.

		if(FreeClass(SoundClass) == FAILED)
			Result = FALSE;
		else
		{
			SoundClass = NULL;

			CloseLibrary(SuperClassBase);
			SuperClassBase = NULL;
		}
	}

	return(Result);
}
