/*
**	dispatch.c - dispatcher for REKO DataType class
**	Copyright  1995-96 Michael Letowski
*/

#define USE_ASM_ROUTINES

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <graphics/displayinfo.h>
#include <graphics/gfx.h>
#include <graphics/modeid.h>
#include <intuition/classes.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/pictureclass.h>
#include <libraries/iffparse.h>
#include <utility/hooks.h>
#include <support/types.h>
#include <support/graphics.h>

#include <string.h>

#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/datatypes.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <proto/utility.h>
#include <clib/support_protos.h>

#include "classbase.h"
#include "dispatch.h"
#ifdef USE_ASM_ROUTINES
#include "card.h"
#endif	/* USE_ASM_ROUTINES */


/****** reko.datatype/reko.datatype *****************************************
*
*   NAME
*       reko.datatype -- data type for REKO Productions' Klondike cardsets.
*
*   FUNCTION
*       The reko data type, a sub-class of the picture.datatype, is used
*       to load (and display) cardsets made for REKO Productions' Klondike
*       game. The data type recognizes type I (55 cards), type II (59 cards)
*       ant type III (68 cards) cardsets, but it can deal with any future
*       cardset. Both HAM(8) and color cardsets are handled.
*
*   METHODS
*       OM_NEW -- Create a new picture object from a REKO cardset. The
*                 source must be a file.
*
*   BUGS
*       HAM8 cardsets will not be displayed on non-AGA machines.
*
*   SEE ALSO
*       picture.datatype.
*
*****************************************************************************
*
*/


/*
**	Private structures
*/
/* Desribes file header */
struct RekoHeader
{
	ULONG rh_ID;
	ULONG rh_BodySize;
	ULONG rh_SingleSize;
	UWORD rh_Height;
	UWORD rh_Width;
	ULONG rh_ModeID;
	UBYTE rh_Depth;
	UBYTE rh_CardsCnt;
};	/* RekoHeader */


/*
**	Private constants
*/
#define REKO_ID				MAKE_ID('R','E','K','O')
#define HEADER_SIZE		sizeof(struct RekoHeader)

#define HCNT					13												/* Number of cards horizontally */
#define VCNT					4													/* Number of cards vertically */
#define REKO_I				55												/* Cards in REKO I cardset */
#define REKO_II				59												/* Cards in REKO II cardset */
#define REKO_III			68												/* Cards in REKO III cardset */

#define BM_FLAGS			(BMF_CLEAR | BMF_INTERLEAVED | BMF_DISPLAYABLE)


/*
**	Private functions prototypes
*/
STATIC ASM Object *Dispatch(R_A0 Class *cl, R_A2 Object *o, R_A1 Msg msg);
STATIC LBOOL GetREKO(struct ClassBase *cb, Object *o, struct TagItem *attrs);

STATIC ULONG GetHCnt(ULONG cards);
#ifndef USE_ASM_ROUTINES
STATIC VOID CopyCard(struct BitMap *bm, MPTR buffer, ULONG x, ULONG y,
											ULONG num, ULONG lSize);
STATIC VOID GetXY(ULONG num, ULONG *x, ULONG *y);
#endif	/* USE_ASM_ROUTINES */


/*
**	Public functions
*/
Class *InitClass(struct ClassBase *cb)
{
	Class *CL;

	/* Create our class (no instance) */
	if(CL=MakeClass(DATATYPENAME,SUPERCLASSNAME,NULL,NULL,0))
	{
		CL->cl_Dispatcher.h_Entry=(HOOKFUNC)Dispatch;
		CL->cl_UserData=(ULONG)cb;
		AddClass(CL);
	}
	return(CL);
}	/* InitClass */


/*
**	Level 1 private functions
*/
STATIC ASM Object *Dispatch(R_A0 Class *cl, R_A2 Object *o, R_A1 Msg msg)
{
	APTR Dummy=(APTR)GetREKO;											/* Trick! - prevent inlining */
	struct ClassBase *cb=(struct ClassBase *)cl->cl_UserData;
	Object *Obj;

	switch(msg->MethodID)
	{
		case OM_NEW:																/* We know this method */
			if(Obj=(Object *)DoSuperMethodA(cl,o,msg))
				unless(GetREKO(cb,Obj,((struct opSet *)msg)->ops_AttrList))
				{
					CoerceMethod(cl,Obj,OM_DISPOSE);
					return(NULL);
				}
			break;
		default:																		/* Let the superclass handle it */
			Obj=(Object *)DoSuperMethodA(cl,o,msg);
			break;
	}
	return(Obj);
}


/*
**	Level 2 private functions
*/
STATIC LBOOL GetREKO(struct ClassBase *cb, Object *o, struct TagItem *attrs)
{
	struct RekoHeader Header;

	struct BitMapHeader *BMHD=NULL;
	struct ColorRegister *CMap=NULL;
	struct BitMap *BM;
	ULONG *CRegs=NULL;

	BPTR FH=0;																		/* File handle */
	MPTR Buffer;																	/* Buffer for reading */
	STRPTR Title;																	/* Picture name */
	ULONG Cards,CWidth,CHeight,LSize,CSize;
	ULONG Width,Height,Depth,NumCol;
	ULONG ModeID;																	/* Display mode */
	ULONG I;

	/* Get default title */
	Title=(STRPTR)GetTagData(DTA_Name,NULL,attrs);

	/* Get file handle and BitMapHeader */
	GetDTAttrs(o,DTA_Handle,&FH,PDTA_BitMapHeader,&BMHD,TAG_DONE);
	try(FH && BMHD,	EXIT)													/* File handle or header not supplied */

	/* Read file header */
	try(Seek(FH,0,OFFSET_BEGINNING)>=0,	EXIT);		/* Position at the beginnig of file */
	try(Read(FH,&Header,HEADER_SIZE)==HEADER_SIZE,	EXIT);

	/* Check header */
	if(Header.rh_ID!=REKO_ID)											/* Not REKO - set error */
		throw2(SetIoErr(ERROR_OBJECT_WRONG_TYPE),	EXIT);

	/* Calculate sizes */
	CWidth=	Header.rh_Width;											/* Get single card's width... */
	CHeight=Header.rh_Height;											/* ... and single card's height */
	Depth=	Header.rh_Depth;											/* Get number of bitplanes */
	NumCol=	1<<(ftst(Header.rh_ModeID,HAM_KEY) ? Depth-2 : Depth);

	Cards=	Header.rh_CardsCnt;										/* Number of cards in set */
	LSize=	CWidth>>3;														/* Width in bytes */
	CSize=	LSize*CHeight*Depth;									/* Card size in bytes */
	if(Cards==0 || CSize==0 || Depth<3 ||					/* Do some internal checks */
			Header.rh_SingleSize!=CSize)
		throw2(SetIoErr(ERROR_OBJECT_WRONG_TYPE),	EXIT);

	/* Fill in BitMapHeader information */
	BMHD->bmh_Width=	BMHD->bmh_PageWidth=	Width=	CWidth*GetHCnt(Cards);
	BMHD->bmh_Height=	BMHD->bmh_PageHeight=	Height=	CHeight*VCNT;
	BMHD->bmh_Depth=	Depth;

	/* Get display mode id */
	ModeID=BestModeID(BIDTAG_SourceID,			Header.rh_ModeID & ~MONITOR_ID_MASK,
										BIDTAG_DesiredWidth,	Width,
										BIDTAG_DesiredHeight,	Height,
										BIDTAG_NominalWidth,	1,
										BIDTAG_NominalHeight,	1,
										BIDTAG_Depth,					Depth,
										TAG_DONE);
	if(ModeID==INVALID_ID)												/* Disallow invalid ID's */
		throw2(SetIoErr(ERROR_OBJECT_WRONG_TYPE),	EXIT);

	/* Set colors */
	SetDTAttrs(o,NULL,NULL,PDTA_NumColors,NumCol,TAG_DONE);
	GetDTAttrs(o,PDTA_ColorRegisters,&CMap,PDTA_CRegs,&CRegs,TAG_DONE);
	try(CMap && CRegs,										EXIT);
	try(Read(FH,CMap,NumCol*3)==NumCol*3,	EXIT);
	CopyColors(CRegs,CMap,NumCol);								/* Set remap palette */

	/* Prepare bitmap */
	try(BM=AllocBitMap(Width,Height,Depth,BM_FLAGS,NULL),	NO_BM);

	/* Allocate temporary buffer */
	try(Buffer=AllocVec(CSize,MEMF_PUBLIC),	NO_MEM);
	/* Read data in 1-card chunks */
	for(I=0; I<Cards; I++)
	{
		try(Read(FH,Buffer,CSize)==CSize,	BODY_ERR);
		CopyCard(BM,Buffer,I,CHeight,Depth,LSize);
	}
	/* Free temporary buffer */
	FreeVec(Buffer);															/* Free temporary buffer */

	/* Set attributes of destination picture */
	SetDTAttrs(o,NULL,NULL,DTA_ObjName,				Title,
													DTA_NominalHoriz,	Width,
													DTA_NominalVert,	Height,
													PDTA_BitMap,			BM,
													PDTA_ModeID,			ModeID,
													TAG_DONE);
	return(TRUE);

	/* Exceptions */
	catch(BODY_ERR,	);
	catch(NO_MEM,		FreeVec(Buffer));
	catch(NO_BM,		FreeBitMap(BM));
	catch(EXIT,			);
	return(FALSE);
}	/* GetREKO */


/*
**	Level 3 private functions
*/
/* Calculate number of cards horizontally */
STATIC ULONG GetHCnt(ULONG cards)
{
	ULONG HCnt;

	if(cards<=REKO_I)															/* REKO-I cardset */
		HCnt=14;
	elif(cards<=REKO_III)													/* REKO-II or REKO-III cardset */
		HCnt=17;
	else																					/* Unknown cardset */
		HCnt=cards/VCNT+(cards%VCNT>0);
	return(HCnt);
}	/* GetHCnt */

#ifndef USE_ASM_ROUTINES
STATIC VOID CopyColors(ULONG *cRegs, struct ColorRegister *cMap, ULONG numCols)
{
	ULONG I;

	for(I=0; I<numCols; I++)
	{
		cRegs[I*3+0]=Color32(cMap[I].red);
		cRegs[I*3+1]=Color32(cMap[I].green);
		cRegs[I*3+2]=Color32(cMap[I].blue);
	}
}	/* CopyColors */

STATIC VOID CopyCard(struct BitMap *bm, MPTR buffer, ULONG num,
											ULONG h, ULONG d, ULONG lSize)
{
	ULONG X,Y,I,DstOffs,BPR=bm->BytesPerRow;

	GetXY(num,&X,&Y);
	DstOffs=Y*h*BPR+X*lSize;
	while(h--)
	{
		for(I=0; I<d; I++)
		{
			memcpy(bm->Planes[I]+DstOffs,buffer,lSize);
			buffer+=lSize;
		}
		DstOffs+=BPR;
	}
}	/* CopyCard */


/*
**	Level 4 private functions
*/
STATIC VOID GetXY(ULONG num, ULONG *x, ULONG *y)
{
	STATIC CONST UBYTE Mapping[REKO_III][2]=
	{
		{13,1},	{13,2},	{13,3},
		{0,0},	{0,1},	{0,2},	{0,3},	{1,0},	{1,1},	{1,2},	{1,3},
		{2,0},	{2,1},	{2,2},	{2,3},	{3,0},	{3,1},	{3,2},	{3,3},
		{4,0},	{4,1},	{4,2},	{4,3},	{5,0},	{5,1},	{5,2},	{5,3},
		{6,0},	{6,1},	{6,2},	{6,3},	{7,0},	{7,1},	{7,2},	{7,3},
		{8,0},	{8,1},	{8,2},	{8,3},	{9,0},	{9,1},	{9,2},	{9,3},
		{10,0},	{10,1},	{10,2},	{10,3},	{11,0},	{11,1},	{11,2},	{11,3},
		{12,0},	{12,1},	{12,2},	{12,3},
		{13,0},	{14,0},	{15,0},	{16,0},
		{14,1},	{15,1},	{16,1},
		{14,2},	{15,2},	{16,2},
		{14,3},	{15,3},	{16,3}
	};	/* Mapping */

	if(num<REKO_III)
	{
		*x=Mapping[num][0];
		*y=Mapping[num][1];
	}
	else
	{
		*x=num/VCNT;
		*y=num%VCNT;
	}
}	/* GetXY */
#endif	/* USE_ASM_ROUTINES */
